@@ -35,9 +35,6 @@
  *   forward this exception.
  */
 
-/* undefine if you have problems with make_sockets() */
-#define INET_SOCKET_PAIR
-
 #include "common.h"
 #include "prototypes.h"
 
@@ -65,7 +62,6 @@ static void parse_socket_error(CLI *, const char *);
 static void print_cipher(CLI *);
 static void auth_user(CLI *);
 static int connect_local(CLI *);
-static void make_sockets(CLI *, int [2]);
 static int connect_remote(CLI *);
 #ifdef SO_ORIGINAL_DST
 static int connect_transparent(CLI *);
@@ -78,11 +74,12 @@ static void reset(int, char *);
 CLI *alloc_client_session(SERVICE_OPTIONS *opt, int rfd, int wfd) {
     CLI *c;
 
-    c=calloc(1, sizeof(CLI));
+    c=str_alloc(sizeof(CLI));
     if(!c) {
         s_log(LOG_ERR, "Memory allocation failed");
         return NULL;
     }
+    str_detach(c);
     c->opt=opt;
     c->local_rfd.fd=rfd;
     c->local_wfd.fd=wfd;
@@ -104,20 +101,26 @@ void *client(void *arg) {
             if(!c->opt->option.retry)
                 break;
             sleep(1); /* FIXME: not a good idea in ucontext threading */
+            str_stats();
+            str_cleanup();
         }
     } else
         run_client(c);
-    free(c);
+    str_free(c);
 #ifdef DEBUG_STACK_SIZE
     stack_info(0); /* display computed value */
 #endif
+#ifdef USE_UCONTEXT
+    s_log(LOG_DEBUG, "Context %ld closed", ready_head->id);
+#endif
+    str_stats();
+    str_cleanup();
+    /* s_log() is not allowed after str_cleanup() */
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
     _endthread();
 #endif
 #ifdef USE_UCONTEXT
-    s_log(LOG_DEBUG, "Context %ld closed", ready_head->id);
     s_poll_wait(NULL, 0, 0); /* wait on poll() */
-    s_log(LOG_ERR, "INTERNAL ERROR: failed to drop context");
 #endif
     return NULL;
 }
@@ -268,8 +271,17 @@ static void init_ssl(CLI *c) {
     }
     SSL_set_ex_data(c->ssl, cli_index, c); /* for callbacks */
     SSL_set_session_id_context(c->ssl, (unsigned char *)sid_ctx,
-        strlen(sid_ctx));
+        strlen(sid_ctx) );
     if(c->opt->option.client) {
+#ifndef OPENSSL_NO_TLSEXT
+        if(c->opt->host_name) {
+            s_log(LOG_DEBUG, "SNI: host name: %s", c->opt->host_name);
+            if(!SSL_set_tlsext_host_name(c->ssl, c->opt->host_name)) {
+                sslerror("SSL_set_tlsext_host_name");
+                longjmp(c->err, 1);
+            }
+        }
+#endif
         if(c->opt->session) {
             enter_critical_section(CRIT_SESSION);
             SSL_set_session(c->ssl, c->opt->session);
@@ -301,9 +313,9 @@ static void init_ssl(CLI *c) {
 
     while(1) {
 #if OPENSSL_VERSION_NUMBER<0x1000002f
-        /* this critical section is a crude workaround for CVE-2010-3864 */
-        /* see http://www.securityfocus.com/bid/44884 for details */
-        /* NOTE: this critical section also covers callbacks (e.g. OCSP) */
+        /* this critical section is a crude workaround for CVE-2010-3864 *
+         * see http://www.securityfocus.com/bid/44884 for details        *
+         * NOTE: this critical section also covers callbacks (e.g. OCSP) */
         enter_critical_section(CRIT_SSL);
 #endif /* OpenSSL version < 1.0.0b */
         if(c->opt->option.client)
@@ -354,6 +366,9 @@ static void init_ssl(CLI *c) {
         s_log(LOG_INFO, "SSL %s: previous session reused",
             c->opt->option.client ? "connected" : "accepted");
     } else { /* a new session was negotiated */
+#ifdef USE_WIN32
+        win_newcert(c->ssl, c->opt);
+#endif
         if(c->opt->option.client) {
             s_log(LOG_INFO, "SSL connected: new session negotiated");
             enter_critical_section(CRIT_SESSION);
@@ -750,10 +765,12 @@ static void parse_socket_error(CLI *c, const char *text) {
 
 static void print_cipher(CLI *c) { /* print negotiated cipher */
     SSL_CIPHER *cipher;
-    char buf[STRLEN], *i, *j;
+    char *buf, *i, *j;
 
+    if(global_options.debug_level<LOG_INFO) /* performance optimization */
+        return;
     cipher=(SSL_CIPHER *)SSL_get_current_cipher(c->ssl);
-    SSL_CIPHER_description(cipher, buf, STRLEN);
+    buf=SSL_CIPHER_description(cipher, NULL, 0);
     i=j=buf;
     do {
         switch(*i) {
@@ -769,6 +786,7 @@ static void print_cipher(CLI *c) { /* print negotiated cipher */
         }
     } while(*i++);
     s_log(LOG_INFO, "Negotiated ciphers: %s", buf);
+    OPENSSL_free(buf);
 }
 
 static void auth_user(CLI *c) {
@@ -776,7 +794,7 @@ static void auth_user(CLI *c) {
     struct servent *s_ent;    /* structure for getservbyname */
 #endif
     SOCKADDR_UNION ident;     /* IDENT socket name */
-    char name[STRLEN];
+    char *line, *type, *system, *user;
 
     if(!c->opt->username)
         return; /* -u option not specified */
@@ -801,19 +819,46 @@ static void auth_user(CLI *c) {
     fdprintf(c, c->fd, "%u , %u",
         ntohs(c->peer_addr.addr[0].in.sin_port),
         ntohs(c->opt->local_addr.addr[0].in.sin_port));
-    if(fdscanf(c, c->fd, "%*[^:]: USERID :%*[^:]:%s", name)!=1) {
-        s_log(LOG_ERR, "Incorrect data from IDENT server");
-        longjmp(c->err, 1);
-    }
+    line=fdgetline(c, c->fd);
     closesocket(c->fd);
     c->fd=-1; /* avoid double close on cleanup */
-    if(strcmp(name, c->opt->username)) {
-        safestring(name);
+    type=strchr(line, ':');
+    if(!type) {
+        s_log(LOG_ERR, "Malformed IDENT response");
+        str_free(line);
+        longjmp(c->err, 1);
+    }
+    *type++='\0';
+    system=strchr(type, ':');
+    if(!system) {
+        s_log(LOG_ERR, "Malformed IDENT response");
+        str_free(line);
+        longjmp(c->err, 1);
+    }
+    *system++='\0';
+    if(strcmp(type, " USERID ")) {
+        s_log(LOG_ERR, "Incorrect INETD response type");
+        str_free(line);
+        longjmp(c->err, 1);
+    }
+    user=strchr(system, ':');
+    if(!user) {
+        s_log(LOG_ERR, "Malformed IDENT response");
+        str_free(line);
+        longjmp(c->err, 1);
+    }
+    *user++='\0';
+    while(*user==' ') /* skip leading spaces */
+        ++user;
+    if(strcmp(user, c->opt->username)) {
+        safestring(user);
         s_log(LOG_WARNING, "Connection from %s REFUSED by IDENT (user %s)",
-            c->accepted_address, name);
+            c->accepted_address, user);
+        str_free(line);
         longjmp(c->err, 1);
     }
     s_log(LOG_INFO, "IDENT authentication passed");
+    str_free(line);
 }
 
 #if defined(_WIN32_WCE) || defined(__vms)
@@ -832,7 +877,8 @@ static int connect_local(CLI *c) { /* spawn local process */
     PROCESS_INFORMATION pi;
     LPTSTR execname_l, execargs_l;
 
-    make_sockets(c, fd);
+    if(make_sockets(fd))
+        longjmp(c->err, 1);
     memset(&si, 0, sizeof si);
     si.cb=sizeof si;
     si.wShowWindow=SW_HIDE;
@@ -843,8 +889,8 @@ static int connect_local(CLI *c) { /* spawn local process */
     execname_l=str2tstr(c->opt->execname);
     execargs_l=str2tstr(c->opt->execargs);
     CreateProcess(execname_l, execargs_l, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi);
-    free(execname_l);
-    free(execargs_l);
+    str_free(execname_l);
+    str_free(execargs_l);
 
     closesocket(fd[1]);
     CloseHandle(pi.hProcess);
@@ -855,7 +901,7 @@ static int connect_local(CLI *c) { /* spawn local process */
 #else /* standard Unix version */
 
 static int connect_local(CLI *c) { /* spawn local process */
-    char env[3][STRLEN], name[STRLEN], *portname;
+    char *name, *portname;
     int fd[2], pid;
     X509 *peer;
 #ifdef HAVE_PTHREAD_SIGMASK
@@ -863,13 +909,14 @@ static int connect_local(CLI *c) { /* spawn local process */
 #endif
 
     if(c->opt->option.pty) {
-        char tty[STRLEN];
+        char tty[64];
 
         if(pty_allocate(fd, fd+1, tty))
             longjmp(c->err, 1);
         s_log(LOG_DEBUG, "TTY=%s allocated", tty);
     } else
-        make_sockets(c, fd);
+        if(make_sockets(fd))
+            longjmp(c->err, 1);
 
     pid=fork();
     c->pid=(unsigned long)pid;
@@ -881,18 +928,18 @@ static int connect_local(CLI *c) { /* spawn local process */
         longjmp(c->err, 1);
     case  0:    /* child */
         closesocket(fd[0]);
+        set_nonblock(fd[1], 0); /* switch back to blocking mode */
         /* dup2() does not copy FD_CLOEXEC flag */
         dup2(fd[1], 0);
         dup2(fd[1], 1);
         if(!global_options.option.foreground)
             dup2(fd[1], 2);
         closesocket(fd[1]); /* not really needed due to FD_CLOEXEC */
-        safecopy(env[0], "REMOTE_HOST=");
-        safeconcat(env[0], c->accepted_address);
-        portname=strrchr(env[0], ':');
+        name=str_dup(c->accepted_address);
+        portname=strrchr(name, ':');
         if(portname) /* strip the port name */
             *portname='\0';
-        putenv(env[0]);
+        putenv(str_printf("REMOTE_HOST=%s", name));
         if(c->opt->option.transparent_src) {
             putenv("LD_PRELOAD=" LIBDIR "/libstunnel.so");
             /* for Tru64 _RLD_LIST is used instead */
@@ -901,16 +948,12 @@ static int connect_local(CLI *c) { /* spawn local process */
         if(c->ssl) {
             peer=SSL_get_peer_certificate(c->ssl);
             if(peer) {
-                safecopy(env[1], "SSL_CLIENT_DN=");
-                X509_NAME_oneline(X509_get_subject_name(peer), name, STRLEN);
+                name=X509_NAME_oneline(X509_get_subject_name(peer), NULL, 0);
                 safestring(name);
-                safeconcat(env[1], name);
-                putenv(env[1]);
-                safecopy(env[2], "SSL_CLIENT_I_DN=");
-                X509_NAME_oneline(X509_get_issuer_name(peer), name, STRLEN);
+                putenv(str_printf("SSL_CLIENT_DN=%s", name));
+                name=X509_NAME_oneline(X509_get_issuer_name(peer), NULL, 0);
                 safestring(name);
-                safeconcat(env[2], name);
-                putenv(env[2]);
+                putenv(str_printf("SSL_CLIENT_I_DN=%s", name));
                 X509_free(peer);
             }
         }
@@ -930,48 +973,6 @@ static int connect_local(CLI *c) { /* spawn local process */
 
 #endif /* not USE_WIN32 or __vms */
 
-static void make_sockets(CLI *c, int fd[2]) { /* make a pair of connected sockets */
-#ifdef INET_SOCKET_PAIR
-    SOCKADDR_UNION addr;
-    socklen_t addrlen;
-    int s; /* temporary socket awaiting for connection */
-
-    s=s_socket(AF_INET, SOCK_STREAM, 0, 1, "socket#1");
-    if(s<0)
-        longjmp(c->err, 1);
-    fd[1]=s_socket(AF_INET, SOCK_STREAM, 0, 1, "socket#2");
-    if(fd[1]<0)
-        longjmp(c->err, 1);
-    addrlen=sizeof addr;
-    memset(&addr, 0, addrlen);
-    addr.in.sin_family=AF_INET;
-    addr.in.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
-    addr.in.sin_port=htons(0); /* dynamic port allocation */
-    if(bind(s, &addr.sa, addrlen))
-        log_error(LOG_DEBUG, get_last_socket_error(), "bind#1");
-    if(bind(fd[1], &addr.sa, addrlen))
-        log_error(LOG_DEBUG, get_last_socket_error(), "bind#2");
-    if(listen(s, 5)) {
-        sockerror("listen");
-        longjmp(c->err, 1);
-    }
-    if(getsockname(s, &addr.sa, &addrlen)) {
-        sockerror("getsockname");
-        longjmp(c->err, 1);
-    }
-    if(connect(fd[1], &addr.sa, addrlen)) {
-        sockerror("connect");
-        longjmp(c->err, 1);
-    }
-    if((fd[0]=s_accept(s, &addr.sa, &addrlen, 1, "accept"))<0)
-        longjmp(c->err, 1);
-    closesocket(s); /* don't care about the result */
-#else
-    if(s_socketpair(AF_UNIX, SOCK_STREAM, 0, fd, 1, "socketpair"))
-        longjmp(c->err, 1);
-#endif
-}
-
 static int connect_remote(CLI *c) { /* connect remote host */
     SOCKADDR_UNION addr;
     SOCKADDR_LIST resolved_list, *address_list;
@@ -1116,14 +1117,16 @@ static void print_bound_address(CLI *c) {
     SOCKADDR_UNION addr;
     socklen_t addrlen=sizeof addr;
 
+    if(global_options.debug_level<LOG_NOTICE) /* performance optimization */
+        return;
     memset(&addr, 0, addrlen);
     if(getsockname(c->fd, (struct sockaddr *)&addr, &addrlen)) {
         sockerror("getsockname");
-    } else {
-        s_ntop(txt, &addr);
-        s_log(LOG_NOTICE,"Service %s connected remote server from %s",
-            c->opt->servname, txt);
+        return;
     }
+    s_ntop(txt, &addr);
+    s_log(LOG_NOTICE,"Service %s connected remote server from %s",
+        c->opt->servname, txt);
 }
 
 static void reset(int fd, char *txt) {
