@@ -52,10 +52,6 @@
 /* I/O buffer size */
 #define BUFFSIZE 16384
 
-/* length of strings (including the terminating '\0' character) */
-/* it can't be lower than 256 bytes or NTLM authentication will break */
-#define STRLEN 256
-
 /* IP address and TCP port textual representation length */
 #define IPLEN 128
 
@@ -105,15 +101,21 @@ typedef int socklen_t;
 #endif
 
 #ifdef USE_PTHREAD
+#ifndef THREADS
 #define THREADS
+#endif
+#ifndef _REENTRANT
 #define _REENTRANT
+#endif
+#ifndef _THREAD_SAFE
 #define _THREAD_SAFE
+#endif
 #include <pthread.h>
 #endif
 
 /* TCP wrapper */
-#if HAVE_TCPD_H && HAVE_LIBWRAP
-#define USE_LIBWRAP
+#if defined HAVE_TCPD_H && defined HAVE_LIBWRAP
+#define USE_LIBWRAP 1
 #endif
 
 /* must be included before sys/stat.h for Ultrix */
@@ -136,20 +138,6 @@ typedef int socklen_t;
 
 #ifdef USE_WIN32
 
-#ifndef HOST
-#ifdef __MINGW32__
-#define HOST "x86-pc-mingw32-gnu"
-#else
-#ifdef _MSC_VER
-#define _QUOTEME(x) #x
-#define QUOTEME(x) _QUOTEME(x)
-#define HOST "x86-pc-msvc-" ## QUOTEME(_MSC_VER)
-#else
-#define HOST "x86-pc-unknown"
-#endif
-#endif
-#endif
-
 typedef unsigned char u8;
 typedef unsigned short u16;
 typedef unsigned long u32;
@@ -190,6 +178,7 @@ typedef unsigned long u32;
 #define EWOULDBLOCK WSAEWOULDBLOCK
 #define EISCONN WSAEISCONN
 #define EADDRINUSE WSAEADDRINUSE
+#define EOPNOTSUPP WSAEOPNOTSUPP
 
 #ifdef EINVAL
 #undef EINVAL
@@ -242,7 +231,7 @@ typedef unsigned long u32;
 
     /* OpenVMS compatibility */
 #ifdef __vms
-#define libdir "__NA__"
+#define LIBDIR "__NA__"
 #define PIDFILE "SYS$LOGIN:STUNNEL.PID"
 #ifdef __alpha
 #define HOST "alpha-openvms"
@@ -382,6 +371,10 @@ extern char *sys_errlist[];
 #include <openssl/fips_rand.h>
 #endif /* USE_FIPS */
 
+#if OPENSSL_VERSION_NUMBER<0x10000000
+#define OPENSSL_NO_TLSEXT
+#endif /* OpenSSL version < 1.0.0 */
+
 #else /* HAVE_OPENSSL */
 
 #include <lhash.h>
@@ -390,16 +383,12 @@ extern char *sys_errlist[];
 #include <crypto.h>
 #include <md4.h>
 #include <des.h>
+#define OPENSSL_NO_TLSEXT
 
 #endif /* HAVE_OPENSSL */
 
 /**************************************** other defines */
 
-/* safe copy for strings declarated as char[STRLEN] */
-#define safecopy(dst, src) \
-    (dst[STRLEN-1]='\0', strncpy((dst), (src), STRLEN-1))
-#define safeconcat(dst, src) \
-    (dst[STRLEN-1]='\0', strncat((dst), (src), STRLEN-strlen(dst)-1))
 /* change all non-printable characters to '.' */
 #define safestring(s) \
     do {unsigned char *p; for(p=(unsigned char *)(s); *p; p++) \
