@@ -44,7 +44,12 @@
 
 #define MAX_HOSTS 16
 
-typedef enum {LOG_MODE_NONE, LOG_MODE_ERROR, LOG_MODE_FULL} LOG_MODE;
+typedef enum {
+    LOG_MODE_NONE,
+    LOG_MODE_ERROR,
+    LOG_MODE_INFO,
+    LOG_MODE_CONFIGURED
+} LOG_MODE;
 
 typedef union sockaddr_union {
     struct sockaddr sa;
@@ -81,11 +86,6 @@ typedef struct {
     int uid, gid;
 #endif
 
-        /* Win32 specific data for gui.c */
-#if defined(USE_WIN32) && !defined(_WIN32_WCE)
-    char *win32_service;
-#endif
-
         /* logging-support data for log.c */
     int debug_level;                              /* debug level for logging */
 #ifndef USE_WIN32
@@ -110,86 +110,104 @@ typedef struct {
 
 extern GLOBAL_OPTIONS global_options;
 
+typedef struct servername_list_struct SERVERNAME_LIST;/* forward declaration */
+
 typedef struct service_options_struct {
-    SSL_CTX *ctx;                                            /*  SSL context */
-    X509_STORE *revocation_store;             /* cert store for CRL checking */
-#ifdef HAVE_OSSL_ENGINE_H
-    ENGINE *engine;                        /* engine to read the private key */
-#endif
     struct service_options_struct *next;   /* next node in the services list */
+    SSL_CTX *ctx;                                            /*  SSL context */
     char *servname;        /* service name for logging & permission checking */
-    SSL_SESSION *session;                           /* jecently used session */
-    char local_address[IPLEN];             /* dotted-decimal address to bind */
+
+        /* service-specific data for sthreads.c */
 #ifndef USE_FORK
     int stack_size;                            /* stack size for this thread */
 #endif
 
-        /* service-specific data for ctx.c */
+        /* service-specific data for verify.c */
     char *ca_dir;                              /* directory for hashed certs */
     char *ca_file;                       /* file containing bunches of certs */
     char *crl_dir;                              /* directory for hashed CRLs */
     char *crl_file;                       /* file containing bunches of CRLs */
+    int verify_level;
+    X509_STORE *revocation_store;             /* cert store for CRL checking */
+    SOCKADDR_LIST ocsp_addr;
+    char *ocsp_path;
+    unsigned long ocsp_flags;
+
+        /* service-specific data for ctx.c */
     char *cipher_list;
     char *cert;                                             /* cert filename */
     char *key;                               /* pem (priv key/cert) filename */
     long session_timeout;
-    int verify_level;
-    int verify_use_only_my;
-    int curve;
     long ssl_options;
-    SOCKADDR_LIST ocsp_addr;
-    char *ocsp_path;
-    unsigned long ocsp_flags;
     SSL_METHOD *client_method, *server_method;
     SOCKADDR_LIST sessiond_addr;
+    SERVERNAME_LIST *servername_list_head, *servername_list_tail;
+    int curve;
+#ifdef HAVE_OSSL_ENGINE_H
+    ENGINE *engine;                        /* engine to read the private key */
+#endif
 
         /* service-specific data for client.c */
     int fd;        /* file descriptor accepting connections for this service */
-    char *execname; /* program name for local mode */
+    SSL_SESSION *session;                           /* recently used session */
+    char *execname;                           /* program name for local mode */
 #ifdef USE_WIN32
-    char *execargs; /* program arguments for local mode */
+    char *execargs;                      /* program arguments for local mode */
 #else
-    char **execargs; /* program arguments for local mode */
+    char **execargs;                     /* program arguments for local mode */
 #endif
     SOCKADDR_LIST local_addr, remote_addr, source_addr;
     char *username;
     char *remote_address;
-    int timeout_busy; /* maximum waiting for data time */
-    int timeout_close; /* maximum close_notify time */
-    int timeout_connect; /* maximum connect() time */
-    int timeout_idle; /* maximum idle connection time */
-    enum {FAILOVER_RR, FAILOVER_PRIO} failover; /* failover strategy */
+    char *host_name;
+    int timeout_busy;                       /* maximum waiting for data time */
+    int timeout_close;                          /* maximum close_notify time */
+    int timeout_connect;                           /* maximum connect() time */
+    int timeout_idle;                        /* maximum idle connection time */
+    enum {FAILOVER_RR, FAILOVER_PRIO} failover;         /* failover strategy */
 
-        /* protocol name for protocol.c */
+        /* service-specific data for protocol.c */
     char *protocol;
     char *protocol_host;
     char *protocol_username;
     char *protocol_password;
     char *protocol_authentication;
 
+        /* service-specific data for gui.c */
+#ifdef USE_WIN32
+    int section_number;
+#endif
+
         /* on/off switches */
     struct {
+        unsigned int accept:1;          /* endpoint: accept */
         unsigned int client:1;
         unsigned int delayed_lookup:1;
-        unsigned int accept:1;
-        unsigned int remote:1;
-        unsigned int retry:1; /* loop remote+program */
-        unsigned int sessiond:1;
-        unsigned int program:1;
-#ifndef USE_WIN32
-        unsigned int pty:1;
-        unsigned int transparent_src:1;
-        unsigned int transparent_dst:1;
-#endif
-        unsigned int ocsp:1;
 #ifdef USE_LIBWRAP
         unsigned int libwrap:1;
 #endif
+        unsigned int remote:1;          /* endpoint: connect */
+        unsigned int retry:1;           /* loop remote+program */
+        unsigned int sessiond:1;
+        unsigned int program:1;         /* endpoint: exec */
+        unsigned int sni:1;             /* endpoint: sni */
+#ifndef USE_WIN32
+        unsigned int pty:1;
+        unsigned int transparent_src:1;
+        unsigned int transparent_dst:1; /* endpoint: transparent destination */
+#endif
+        unsigned int ocsp:1;
     } option;
 } SERVICE_OPTIONS;
 
 extern SERVICE_OPTIONS service_options;
 
+struct servername_list_struct {
+    char *servername;
+    SERVICE_OPTIONS *opt;
+    struct servername_list_struct *next;
+};
+
 typedef enum {
     TYPE_NONE, TYPE_FLAG, TYPE_INT, TYPE_LINGER, TYPE_TIMEVAL, TYPE_STRING
 } VAL_TYPE;
@@ -244,10 +262,11 @@ typedef struct disk_file {
 extern volatile int num_clients;
 
 void main_initialize(char *, char *);
-void main_execute(void);
+void daemon_loop(void);
+void unbind_ports(void);
 int bind_ports(void);
 #if !defined (USE_WIN32) && !defined (__vms) && !defined(USE_OS2)
-void drop_privileges(void);
+void drop_privileges(int);
 #endif
 int s_socket(int, int, int, int, char *);
 int s_pipe(int [2], int, char *);
@@ -255,6 +274,7 @@ int s_socketpair(int, int, int, int [2], int, char *);
 int s_accept(int, struct sockaddr *, socklen_t *, int, char *);
 void stunnel_info(int);
 void die(int);
+void set_nonblock(int, unsigned long);
 
 /**************************************** prototypes for log.c */
 
@@ -296,7 +316,7 @@ ENGINE *get_engine(int);
 /**************************************** prototypes for options.c */
 
 void parse_commandline(char *, char *);
-void parse_conf(char *, CONF_TYPE);
+int parse_conf(char *, CONF_TYPE);
 
 /**************************************** prototypes for ctx.c */
 
@@ -315,14 +335,26 @@ int s_poll_canread(s_poll_set *, int);
 int s_poll_canwrite(s_poll_set *, int);
 int s_poll_error(s_poll_set *, int);
 int s_poll_wait(s_poll_set *, int, int);
-#if !defined(USE_WIN32) && !defined(USE_OS2)
+
+#ifdef USE_WIN32
+#define SIGNAL_RELOAD_CONFIG    1
+#define SIGNAL_REOPEN_LOG       2
+#define SIGNAL_TERMINATE        3
+#else
+#define SIGNAL_RELOAD_CONFIG    SIGHUP
+#define SIGNAL_REOPEN_LOG       SIGUSR1
+#define SIGNAL_TERMINATE        SIGTERM
+#endif
 void signal_handler(int);
 int signal_pipe_init(void);
+void signal_post(int);
+#if !defined(USE_WIN32) && !defined(USE_OS2)
 void child_status(void);  /* dead libwrap or 'exec' process detected */
 #endif
+
 int set_socket_options(int, int);
-void set_nonblock(int, unsigned long);
 int get_socket_error(const int);
+int make_sockets(int [2]);
 
 /**************************************** prototypes for client.c */
 
@@ -332,24 +364,26 @@ typedef struct {
 } FD;
 
 typedef struct {
+    SSL *ssl; /* SSL connnection */
     SERVICE_OPTIONS *opt;
-    char accepted_address[IPLEN]; /* text */
+    jmp_buf err; /* exception handler */
+
+    char accepted_address[IPLEN]; /* IP address as text for logging */
     SOCKADDR_LIST peer_addr; /* peer address */
     FD local_rfd, local_wfd; /* read and write local descriptors */
     FD remote_fd; /* remote file descriptor */
-    SSL *ssl; /* SSL Connection */
     SOCKADDR_LIST bind_addr;
         /* IP for explicit local bind or transparent proxy */
-    unsigned long pid; /* PID of local process */
+    unsigned long pid; /* PID of the local process */
     int fd; /* temporary file descriptor */
-    jmp_buf err;
 
+    /* data for transfer() function */
     char sock_buff[BUFFSIZE]; /* socket read buffer */
     char ssl_buff[BUFFSIZE]; /* SSL read buffer */
     int sock_ptr, ssl_ptr; /* index of first unused byte in buffer */
     FD *sock_rfd, *sock_wfd; /* read and write socket descriptors */
     FD *ssl_rfd, *ssl_wfd; /* read and write SSL descriptors */
-    int sock_bytes, ssl_bytes; /* bytes written to socket and ssl */
+    int sock_bytes, ssl_bytes; /* bytes written to socket and SSL */
     s_poll_set fds; /* file descriptors */
 } CLI;
 
@@ -362,20 +396,14 @@ int connect_blocking(CLI *, SOCKADDR_UNION *, socklen_t);
 void write_blocking(CLI *, int fd, void *, int);
 void read_blocking(CLI *, int fd, void *, int);
 void fdputline(CLI *, int, const char *);
-void fdgetline(CLI *, int, char *);
+char *fdgetline(CLI *, int);
 /* descriptor versions of fprintf/fscanf */
-int fdprintf(CLI *, int, const char *, ...)
+void fdprintf(CLI *, int, const char *, ...)
 #ifdef __GNUC__
        __attribute__ ((format (printf, 3, 4)));
 #else
        ;
 #endif
-int fdscanf(CLI *, int, const char *, char *)
-#ifdef __GNUC__
-       __attribute__ ((format (scanf, 3, 0)));
-#else
-       ;
-#endif
 
 /**************************************** prototype for protocol.c */
 
@@ -390,12 +418,17 @@ char *s_ntop(char *, SOCKADDR_UNION *);
 /**************************************** prototypes for sthreads.c */
 
 typedef enum {
-    CRIT_KEYGEN, CRIT_INET, CRIT_CLIENTS,
-    CRIT_WIN_LOG, CRIT_SESSION, CRIT_LIBWRAP,
+    CRIT_CLIENTS, CRIT_SESSION, /* client.c */
 #if OPENSSL_VERSION_NUMBER<0x1000002f
-    CRIT_SSL,
+    CRIT_SSL,                   /* client.c */
 #endif /* OpenSSL version < 1.0.0b */
-    CRIT_SECTIONS
+    CRIT_INET,                  /* resolver.c */
+#ifdef USE_WIN32
+    CRIT_WIN_LOG,               /* gui.c */
+#else
+    CRIT_LIBWRAP,               /* libwrap.c */
+#endif
+    CRIT_SECTIONS               /* number of critical sections */
 } SECTION_CODE;
 
 void enter_critical_section(SECTION_CODE);
@@ -413,6 +446,7 @@ typedef struct CONTEXT_STRUCTURE {
     int ready; /* number of ready file descriptors */
     time_t finish; /* when to finish poll() for this context */
     struct CONTEXT_STRUCTURE *next; /* next context on a list */
+    void *tls; /* thread local storage for str.c */
 } CONTEXT;
 extern CONTEXT *ready_head, *ready_tail;
 extern CONTEXT *waiting_head, *waiting_tail;
@@ -434,7 +468,9 @@ typedef struct {
 
 #ifdef USE_WIN32
 void win_log(char *);
-void exit_win32(int);
+void win_exit(int);
+void win_newconfig(int);
+void win_newcert(SSL *, SERVICE_OPTIONS *);
 int passwd_cb(char *, int, int, void *);
 #ifdef HAVE_OSSL_ENGINE_H
 int pin_cb(UI *, UI_STRING *);
@@ -472,6 +508,28 @@ LPSTR tstr2str(const LPTSTR);
 void libwrap_init(int);
 void libwrap_auth(CLI *);
 
+/**************************************** prototypes for str.c */
+
+void str_init();
+void str_canary();
+void str_cleanup();
+void str_stats();
+void *str_alloc(size_t);
+void *str_realloc_debug(void *, size_t, char *, int);
+#define str_realloc(a, b) str_realloc_debug((a), (b), __FILE__, __LINE__)
+void str_detach_debug(void *, char *, int);
+#define str_detach(a) str_detach_debug((a), __FILE__, __LINE__)
+void str_free_debug(void *, char *, int);
+#define str_free(a) str_free_debug((a), __FILE__, __LINE__)
+char *str_dup(const char *);
+char *str_vprintf(const char *, va_list);
+char *str_printf(const char *, ...)
+#ifdef __GNUC__
+    __attribute__ ((format (printf, 1, 2)));
+#else
+    ;
+#endif
+
 #endif /* defined PROTOTYPES_H */
 
 /* end of prototypes.h */
